package Rose::DB::Object::MixIn;

use strict;

use base 'Rose::Object::MixIn';

our $VERSION = '0.764';

1;

__END__

=head1 NAME

Rose::DB::Object::MixIn - A base class for mix-ins.

=head1 SYNOPSIS

  # This module is deprecated.  Please use Rose::Object::MixIn instead.

=head1 DESCRIPTION

This class is a trivial subclass of L<Rose::Object::MixIn>.  It exists for backward compatibility reasons only.  Please see L<Rose::Object::MixIn> for full documentation, and use that class instead of this one.

=head1 AUTHOR

John C. Siracusa (siracusa@gmail.com)

=head1 LICENSE

Copyright (c) 2010 by John C. Siracusa.  All rights reserved.  This program is
free software; you can redistribute it and/or modify it under the same terms
as Perl itself.
